import sys

from twisted.internet.protocol import Protocol, Factory
from twisted.internet.endpoints import (
    # the "TCP4" names are misleading, they should be called "TCP4or6"
    TCP4ServerEndpoint, TCP4ClientEndpoint, connectProtocol)
from twisted.internet import reactor

class OneWayProxySource(Protocol):
    def connectionMade(self):
        self.factory.port.stopListening()
        dest_endpoint = TCP4ClientEndpoint(reactor,
                                           self.factory.dest_host,
                                           self.factory.dest_port)
        self.dest_protocol = OneWayProxyDest()
        d = connectProtocol(dest_endpoint, self.dest_protocol)
        d.addCallback(lambda p: p.connection_made())
        def failed_connect(failure):
            print("Failed to connect:")
            failure.printTraceBack()
            reactor.stop()
        d.addErrback(failed_connect)

    def dataReceived(self, data):
        self.dest_protocol.send_data(data)

    def connectionLost(self, reason):
        self.dest_protocol.close()

class OneWayProxySourceFactory(Factory):
    protocol = OneWayProxySource

    def __init__(self, dest_host, dest_port):
        self.dest_host = dest_host
        self.dest_port = dest_port

class OneWayProxyDest(Protocol):
    def __init__(self):
        self._buffer = bytearray()
        self._close_buffered = False
        self._connected = False

    def send_data(self, data):
        assert not self._close_buffered
        if not self._connected:
            self._buffer.extend(data)
        else:
            self.transport.write(data)

    def close(self):
        if not self._connected:
            self._close_buffered = True
        else:
            self.transport.loseConnection()

    def connection_made(self):
        # Connection made -- flush buffered data
        self.transport.write(self._buffer)
        del self._buffer
        if self._close_buffered:
            self.transport.loseConnection()
        self._connected = True

    def connectionLost(self, reason):
        reactor.stop()

def main(source_port, dest_host, dest_port):
    factory = OneWayProxySourceFactory(dest_host, dest_port)
    endpoint = TCP4ServerEndpoint(reactor, source_port)
    d = endpoint.listen(factory)
    def save_port(port):
        factory.port = port
    d.addCallback(save_port)
    def listening_fail(failure):
        print("Well, that didn't work: {}".format(failure))
        reactor.stop()
    reactor.run()

if __name__ == "__main__":
    try:
        args = [int(sys.argv[1]), sys.argv[2], int(sys.argv[3])]
    except Exception:
        print("Usage: {} SOURCE_PORT DEST_HOST DEST_PORT".format(__file__))
    else:
        main(*args)
